"""Submit a job to a remote server and handle logging.

This module can be used either with Celery, in which case it will run in a
background thread, or as a normal function call, in which case it will block
the current execution thread.

"""
from __future__ import absolute_import, unicode_literals, print_function
import os
import sys
import logging

logger = logging.getLogger(__name__)

import io
import os.path
import select
import socket
import collections
import fnmatch

import six
from paramiko import (
    AuthenticationException, BadHostKeyException, AuthenticationException,
)
from paramiko.client import SSHClient, AutoAddPolicy
from threading import Thread
from django.core.files import File

from .models import Log, Job, Result, Interpreter
from .remote import RemoteWrapper

try:
    from celery import shared_task
except ImportError:
    logger.info('Could not import Celery. '
                'Tasks will not be implemented by Celery\'s queue.')

    def shared_task(func):
        """Naive wrapper in case Celery does not exist."""
        def delay(*args, **kwargs):
            return func(*args, **kwargs)

        func.delay = delay
        return func


class LogPolicy(object):
    """Specify how logging should be done when running a job."""

    LOG_NONE = 0
    """Don't log anything from the running job."""

    LOG_LIVE = 1
    """Create Log objects immediately when they are received."""

    LOG_TOTAL = 2
    """Combine all of stdout and stderr at the end of the job."""


def is_matching(filename, patterns=None):
    """Check if a filename matches the list of positive and negative patterns.

    Positive patterns are strings like ``"1.txt"``, ``"[23].txt"``, or
    ``"*.txt"``.

    Negative patterns are strings like ``"!1.txt"``, ``"![23].txt"``, or
    ``"!*.txt"``.

    Each pattern is checked in turn, so the list of patterns ``["!*.txt",
    "1.txt"]`` will still match ``"1.txt"``.

    >>> from django_remote_submission.tasks import is_matching
    >>> is_matching("1.txt", patterns=["1.txt"])
    True
    >>> is_matching("1.txt", patterns=["[12].txt"])
    True
    >>> is_matching("1.txt", patterns=["*.txt"])
    True
    >>> is_matching("1.txt", patterns=["1.txt", "!*.txt"])
    False
    >>> is_matching("1.txt", patterns=["!*.txt", "[12].txt"])
    True

    """
    if patterns is None:
        patterns = ['*']

    is_matching = False

    for pattern in patterns:
        if not pattern.startswith('!'):
            if fnmatch.fnmatch(filename, pattern):
                is_matching = True
        else:
            if fnmatch.fnmatch(filename, pattern[1:]):
                is_matching = False

    return is_matching


class LogContainer(object):
    """Manage logs sent by a job according to the log policy.

    .. testsetup::

       from django_remote_submission.models import Job, Server, Interpreter
       from django.contrib.auth import get_user_model
       python3 = Interpreter(name='Python 3', path='/bin/python3', arguments=['-u'])
       server = Server(title='Remote', hostname='foo.invalid', port=22)
       user = get_user_model()(username='john')
       job = Job(title='My Job', program='print("hello world")',
           remote_directory='/tmp/', remote_filename='foobar.py',
           owner=user, server=server, interpreter=python3,
       )

    >>> from django_remote_submission.tasks import LogContainer, LogPolicy
    >>> from datetime import datetime
    >>> now = datetime(year=2017, month=1, day=2, hour=3, minute=4, second=5)
    >>> logs = LogContainer(job, LogPolicy.LOG_LIVE)
    >>> logs.write_stdout(now, 'hello world')  # doctest: +SKIP
    >>> Log.objects.get()  # doctest: +SKIP
    <Log: 2017-01-02 03:04:05 My Job>

    """

    LogLine = collections.namedtuple('LogLine', [
        'now', 'output',
    ])

    def __init__(self, job, log_policy):
        """Instantiate a log container.

        :param models.Job job: the job these logs are coming from
        :param LogPolicy log_policy: the policy to use for logging

        """
        self.job = job
        """The job that these logs are coming from."""

        self.log_policy = log_policy
        """The policy to use when logging."""

        self._stdout = []
        """The list of log lines that came from stdout."""

        self._stderr = []
        """The list of log lines that came from stderr."""

    def _write(self, lst, now, output):
        """Append the current log entry to the given list and flush.

        :param lst: either :attr:`stdout` or :attr:`stderr`
        :param datetime.datetime now: the time this line was produced
        :param str output: the line of output from the job

        """
        if self.log_policy != LogPolicy.LOG_NONE:
            lst.append(LogContainer.LogLine(
                now=now,
                output=output,
            ))

        if self.log_policy == LogPolicy.LOG_LIVE:
            self.flush()

    def write_stdout(self, now, output):
        """Write some output from a job's stdout stream.

        :param datetime.datetime now: the time this output was produced
        :param str output: the output that was produced

        """
        self._write(self._stdout, now, output)

    def write_stderr(self, now, output):
        """Write some output from a job's stderr stream.

        :param datetime.datetime now: the time this output was produced
        :param str output: the output that was produced

        """
        self._write(self._stderr, now, output)

    def flush(self):
        """Flush the stdout and stderr lists to Django models.

        If the :attr:`log_policy` is :const:`LogPolicy.LOG_TOTAL`, this method
        will need to be called at the end of the job to ensure all the data
        gets written out.

        There is no penalty for calling this method multiple times, so it can
        be called at the end of the job regardless of which log policy is used.

        """
        if len(self._stdout) > 0:
            Log.objects.create(
                time=self._stdout[-1].now,
                content=''.join(line.output for line in self._stdout),
                stream='stdout',
                job=self.job,
            )

            del self._stdout[:]

        if len(self._stderr) > 0:
            Log.objects.create(
                time=self._stderr[-1].now,
                content=''.join(line.output for line in self._stderr),
                stream='stderr',
                job=self.job,
            )

            del self._stderr[:]


@shared_task
def submit_job_to_server(job_pk, password, username=None, timeout=None,
                         log_policy=LogPolicy.LOG_LIVE, store_results=None,
                         wrapper_cls=RemoteWrapper):
    """Submit a job to the remote server.

    This can be used as a Celery task, if the library is installed and running.

    :param int job_pk: the primary key of the :class:`models.Job` to submit
    :param str password: the password of the user submitting the job
    :param str username: the username of the user submitting, if it is
        different from the owner of the job
    :param datetime.timedelta timeout: the timeout for running the job
    :param LogPolicy log_policy: the policy to use for logging
    :param list(str) store_results: the patterns to use for the results to store
    :param class wrapper_cls: the class constructor to connect to a remote host

    """
    job = Job.objects.get(pk=job_pk)

    if username is None:
        username = job.owner.username

    wrapper = wrapper_cls(
        hostname=job.server.hostname,
        username=username,
        port=job.server.port,
    )

    logs = LogContainer(
        job=job,
        log_policy=log_policy,
    )

    with wrapper.connect(password):
        wrapper.chdir(job.remote_directory)

        with wrapper.open(job.remote_filename, 'wt') as f:
            f.write(job.program)

        import time; time.sleep(1)

        job.status = Job.STATUS.submitted
        job.save()

        interp = job.interpreter.path
        workdir = job.remote_directory
        args = job.interpreter.arguments
        filename = job.remote_filename

        job_status = wrapper.exec_command(
            [interp] + args + [filename],
            workdir,
            timeout=timeout,
            stdout_handler=logs.write_stdout,
            stderr_handler=logs.write_stderr,
        )

        logs.flush()

        job.status = Job.STATUS.success if job_status else Job.STATUS.failure
        job.save()

        file_attrs = wrapper.listdir_attr()
        file_map = { attr.filename: attr for attr in file_attrs }
        script_attr = file_map[job.remote_filename]
        script_mtime = script_attr.st_mtime

        results = []
        for attr in file_attrs:
            print('{!r}'.format(attr))

            if attr is script_attr:
                continue

            if attr.st_mtime < script_mtime:
                continue

            if not is_matching(attr.filename, store_results):
                print('is_matching: {}'.format(attr.filename))
                continue
            else:
                print('not is_matching: {}'.format(attr.filename))

            result = Result.objects.create(
                remote_filename=attr.filename,
                job=job,
            )

            with wrapper.open(attr.filename, 'rb') as f:
                result.local_file.save(attr.filename, File(f), save=True)

            results.append(result)

    return { r.remote_filename: r.pk for r in results }



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-01-20 02:38
from __future__ import unicode_literals

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('django_remote_submission', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='job',
            name='uuid',
            field=models.UUIDField(default=uuid.uuid4, editable=False, help_text='A unique identifier for use in grouping Result files', verbose_name='Job UUID'),
        ),
    ]



"""Provide an admin interface for managing jobs."""

# -*- coding: utf-8 -*-
from django.contrib import admin
from django import forms
from django.utils.translation import ungettext_lazy as _
from django.shortcuts import render
from django.http.response import HttpResponseRedirect

from .models import Server, Job, Log, Interpreter, Result
from .tasks import submit_job_to_server

@admin.register(Interpreter)
class InterpreterAdmin(admin.ModelAdmin):
    """Manage interpreters with default admin interface."""
    pass


@admin.register(Result)
class ResultAdmin(admin.ModelAdmin):
    """Manage Results with default admin interface."""
    pass


@admin.register(Server)
class ServerAdmin(admin.ModelAdmin):
    """Manage servers and allow selecting interpreters from available ones."""

    filter_horizontal = ('interpreters',)


@admin.register(Job)
class JobAdmin(admin.ModelAdmin):
    """Manage jobs with ability to submit the job from the interface.

    To submit a job:

    1. Make sure it is created and has all the correct fields, including the
    program, server, and interpreter.

    2. Go back to the main job index.

    3. Select the job.

    4. Select "Submit to server" from the actions list.

    5. Put in the username of the person to execute the job (or blank if it's
    the same as the owner's Django username).

    6. Put in the password of the user to execute the job.

    7. Click "Submit to Server".

    """

    actions = ['submit_to_server']

    class RequestPasswordForm(forms.Form):
        """Provide a form to put in the username and password of job's owner.

        The hidden inputs are requried to work with the action list in the
        admin interface. These were hard to find because having a second page
        for actions in Django isn't very well documented, and information is
        spread through a handful of blog posts.

        """

        _selected_action = forms.CharField(
            widget=forms.MultipleHiddenInput,
            required=False,
        )
        action = forms.CharField(
            widget=forms.HiddenInput,
            required=False,
        )
        select_across = forms.BooleanField(
            widget=forms.HiddenInput,
            required=False,
        )

        username = forms.CharField(required=False)
        password = forms.CharField(widget=forms.PasswordInput)

    def submit_to_server(self, request, queryset):
        """Submit job to server via an admin interface action."""
        form = None

        if 'apply' in request.POST:
            form = JobAdmin.RequestPasswordForm(request.POST)

            if form.is_valid():
                password = form.cleaned_data['password']
                username = form.cleaned_data['username']

                count = 0
                for job in queryset:
                    submit_job_to_server.delay(
                        job_pk=job.pk,
                        password=password,
                        username=username,
                    )
                    count += 1

                message = _(
                    'Successfully submitted %(count)s job',
                    'Successfully submitted %(count)s jobs',
                    count,
                ) % {
                    'count': count,
                }

                self.message_user(request, message)
                return HttpResponseRedirect(request.get_full_path())

        if not form:
            # This part is required to make the intermediate page work with the
            # Django actions.
            selected_action = request.POST.getlist(admin.ACTION_CHECKBOX_NAME)
            index = int(request.POST.get('index', 0))
            form = JobAdmin.RequestPasswordForm(initial={
                admin.ACTION_CHECKBOX_NAME: selected_action,
                'select_across': request.POST['select_across'],
                'action': request.POST['action'],
            })

        return render(
            request,
            'admin/django_remote_submission/submit_to_server.html',
            {
                'jobs': queryset,
                'form': form,
            },
        )


@admin.register(Log)
class LogAdmin(admin.ModelAdmin):
    """Manage logs with the default admin interface."""

    pass



"""Provide the forms for working with this app's models."""

# -*- coding: utf-8 -*-
from django import forms

from .models import Server, Job


class ServerForm(forms.ModelForm):
    """Provide a form for inputting information about a remote server."""

    class Meta:  # noqa: D101
        model = Server
        fields = ('title', 'hostname')


class JobForm(forms.ModelForm):
    """Provide a form for inputting information about a job."""

    class Meta:  # noqa: D101
        model = Job
        fields = ('title', 'program', 'status', 'server')



"""Provides a wrapper around Paramiko to simplify the API.

This module is meant to be a general wrapper so that a ``LocalWrapper`` can
also be created to run tests in continuous integration services where SSH is
not available.

"""

from __future__ import absolute_import, unicode_literals, print_function
import logging
import os
import textwrap
import datetime
import uuid

try:
    from shlex import quote as cmd_quote
except ImportError:
    from pipes import quote as cmd_quote

from paramiko import (
    AuthenticationException, BadHostKeyException,
)
from paramiko.client import SSHClient, AutoAddPolicy
from django.utils.timezone import now
import six

logger = logging.getLogger(__name__)


class RemoteWrapper(object):
    """Wrapper around Paramiko which simplifies the remote connection API.

    The goal with this class is to also be able to provide a ``LocalWrapper``
    which works with the local file system, so that tests can be run on
    continuous integration servers.

    """

    def __init__(self, hostname, username, port=22):
        """Initialize the wrapper.

        :param str hostname: the hostname of the server to connect to
        :param str username: the username of the user on the remote server
        :param int port: the SSH port to connect to

        """
        self.hostname = hostname
        self.username = username
        self.port = port

        self._client = None
        """The Paramiko Client instance"""

        self._sftp = None
        """The Paramiko SFTP instance"""

    def __enter__(self):
        """Allow the use of ``with wrapper:``."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Allow the use of ``with wrapper:``."""
        self.close()

    def connect(self, password=None, public_key_filename=None):
        """Connect to the remote host with the given password and public key.

        Meant to be used like::

            with wrapper.connect(password='password0'):
                pass

        :param str password: the password of the user on the remote server
        :param str public_key_filename: the file containing the public key

        """
        if public_key_filename is None:
            public_key_filename = os.path.expanduser('~/.ssh/id_rsa.pub')

        self._public_key_filename = public_key_filename
        self._client = self._start_client(password, public_key_filename)
        self._sftp = self._client.open_sftp()
        return self

    def close(self):
        """Close any open connections and clear their attributes."""
        self._sftp.close()
        self._sftp = None

        self._client.close()
        self._client = None

    def chdir(self, remote_directory):
        """Change directories to the remote directory.

        :param str remote_directory: the directory to change to

        """
        self._sftp.chdir(remote_directory)

    def open(self, filename, mode):
        """Open a file from the last used remote directory.

        :param str filename: the name of the file to open
        :param str mode: the mode to use to open the file (see :func:`file`'s
            documentation for more information)

        """
        return self._sftp.open(filename, mode)

    def listdir_attr(self):
        """Retrieve a list of files and their attributes.

        Each object is guaranteed to have a ``filename`` attribute as well as
        an ``st_mtime`` attribute, which gives the last modified time in
        seconds.

        """
        return self._sftp.listdir_attr()

    def delete_key(self):
        """Delete the server's public key from remote host.

        For example::

            wrapper = RemoteWrapper(hostname, username)
            with wrapper.connect(password, public_key_filename):
                wrapper.delete_key()

        """
        if self._client is None:
            raise ValueError('Wrapper must be connected before delete_key is called')

        with open(self._public_key_filename, 'rt') as f:
            key = f.read().strip()

        self.chdir('/tmp')

        filename = 'django-remote-submission-{}'.format(uuid.uuid4)
        with self.open(filename, 'wt') as f:
            program = textwrap.dedent('''\
            sed -i.bak -e /{key}/d $HOME/.ssh/authorized_keys
            '''.format(key=cmd_quote(key.replace('/', '\/'))))

            f.write(program)

        args = [
            'bash', '/tmp/' + filename,
        ]

        self.exec_command(args, '/')


    def exec_command(self, args, workdir, timeout=None, stdout_handler=None,
                     stderr_handler=None):
        """Execute a command on the remote server.

        An example of how to use this function::

            from datetime import timedelta
            wrapper.exec_command(
                args=["ls", "-la", "."],
                workdir="/",
                timeout=timedelta(minute=5),
                stdout_handler=lambda now, output: print('stdout, now, output),
                stderr_handler=lambda now, output: print('stderr, now, output),
            )

        :param list(str) args: the command and arguments to run
        :param str workdir: the directory to run the commands from
        :param datetime.timedelta timeout: the timeout to use for the command
        :param stdout_handler: a function that accepts ``now`` and ``output``
            parameters and is called when new output appears on stdout.
        :param stderr_handler: a function that accepts ``now`` and ``output``
            parameters and is called when new output appears on stderr.

        """
        chdir = self._make_command(['cd', workdir], None)
        run = self._make_command(args, timeout)
        command = '{} && {}'.format(chdir, run)
        print('exec_command(command={!r})'.format(command))

        transport = self._client.get_transport()
        channel = transport.open_session()

        channel.exec_command(command)
        while True:
            current_time = now()
            if channel.recv_ready():
                output = channel.recv(1024).decode('utf-8')
                if stdout_handler is not None:
                    stdout_handler(current_time, output)

            if channel.recv_stderr_ready():
                output = channel.recv_stderr(1024).decode('utf-8')
                if stderr_handler is not None:
                    stderr_handler(current_time, output)

            if channel.exit_status_ready():
                if channel.recv_ready() or channel.recv_stderr_ready():
                    continue

                if channel.recv_exit_status() == 0:
                    return True
                else:
                    return False

    def _start_client(self, password, public_key_filename):
        client = SSHClient()
        client.set_missing_host_key_policy(AutoAddPolicy())

        server_hostname = self.hostname
        server_port = self.port

        try:
            logger.info("Connecting to %s with public key.", server_hostname)
            client.connect(
                server_hostname,
                port=server_port,
                username=self.username,
                key_filename=public_key_filename,
            )

        except (AuthenticationException, BadHostKeyException) as e:
            if password is None:
                logger.error("Connection with public key failed! "
                             "The password is mandatory")
                six.raise_from(ValueError('needs password'), e)

            try:
                logger.info("Connecting to %s with password.", server_hostname)
                client.connect(
                    server_hostname,
                    port=server_port,
                    username=self.username,
                    password=password,
                )
                deploy_key_if_it_doesnt_exist(client, public_key_filename)

            except AuthenticationException as e:
                logger.error("Authenctication error! Wrong password...")
                six.raise_from(ValueError('incorrect password'), e)

        return client

    def _make_command(self, args, timeout):
        command = ' '.join(cmd_quote(arg) for arg in args)

        if timeout is not None:
            command = 'timeout {}s {}'.format(timeout.total_seconds(), command)

        return command


def deploy_key_if_it_doesnt_exist(client, public_key_filename):
    """Deploy our public key to the remote server.

    :param paramiko.client.SSHClient client: an existing Paramiko client
    :param str public_key_filename: the name of the file with the public key

    """
    with open(public_key_filename, 'rt') as f:
        key = f.read()

    client.exec_command('mkdir -p ~/.ssh/')
    client.exec_command('chmod 700 ~/.ssh/')
    client.exec_command('chmod 644 ~/.ssh/authorized_keys')

    command = textwrap.dedent('''\
    KEY={}
    if [ -z "$(grep \"$KEY\" ~/.ssh/authorized_keys )" ]; then
        printf $'%s\n' "$KEY" >> ~/.ssh/authorized_keys
        echo key added.
    fi
    '''.format(cmd_quote(key.strip('\n'))))

    stdin, stdout, stderr = client.exec_command(command)
    logger.debug(stdout.readlines())
    logger.debug(stderr.readlines())

    client.exec_command('chmod 644 ~/.ssh/authorized_keys')
